@@ -1,6 +1,7 @@
 """ Back-end for AWS S3.
 
 Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
+
 """
 
 # Docs: http://boto.readthedocs.org/en/latest/
@@ -12,11 +13,10 @@ if True:  # Imports and constants
         import Store
 
         import boto
+        import collections
         import datetime
         import io
         import logging
-        import os.path
-        import pprint
         import re
         import sys
     if True:  # Constants
@@ -30,37 +30,32 @@ if True:  # Imports and constants
         isEncrypted = True
 
         logger = logging.getLogger(__name__)
-        # logger.setLevel('DEBUG')
+
+# logger.setLevel('DEBUG')
 
 
 class S3Store(Store.Store):
 
     """ An S3 bucket synchronization source or sink. """
 
-    def __init__(self, host, path):
+    def __init__(self, host, path, isDest):
         """ Initialize.
 
         host is the bucket name.
         path is an object key prefix to use.
-        """
-        self.bucketName = host
 
-        if path:
-            path = path.strip("/")
-        if path:
-            path += "/"
-        else:
-            path = ""
-        self.prefix = path
+        """
+        super(S3Store, self).__init__("/" + path, isDest)
+
+        self.isDest = isDest
+
+        self.bucketName = host
 
         self.keyPattern = re.compile(S3Store.theKeyPattern % ())
 
-        # List of dict with from, to, size, path
+        # { fromVol: [diff] }
         self.diffs = None
 
-        # { uuid: { path, uuid, } }
-        self.vols = None
-
         logger.info("Listing %s contents...", self)
 
         try:
@@ -70,7 +65,7 @@ class S3Store(Store.Store):
             raise
 
         self.bucket = s3.get_bucket(self.bucketName)
-        self._listBucket()
+        self._fillVolumesAndPaths()
 
     def __unicode__(self):
         """ Return text description. """
@@ -80,76 +75,69 @@ class S3Store(Store.Store):
         """ Return text description. """
         return unicode(self).encode('utf-8')
 
-    def _listBucket(self):
-        self.vols = {}
-        self.diffs = []
+    def _fillVolumesAndPaths(self):
+        """ Fill in self.paths. """
+        self.diffs = collections.defaultdict((lambda: []))
 
         for key in self.bucket.list():
-            diff = self._parseKeyName(key.name)
+            keyInfo = self._parseKeyName(key.name)
 
-            if diff is None:
-                logger.info("Ignoring '%s' in S3", key.name)
+            if keyInfo is None:
+                if key.name[-1:] != '/':
+                    logger.warn("Ignoring '%s' in S3", key.name)
                 continue
 
-            if diff['from'] == 'None':
-                diff['from'] = None
+            if keyInfo['type'] == 'info':
+                stream = io.BytesIO()
+                key.get_contents_to_file(stream)
+                Store.Volume.readInfo(stream)
+                continue
 
-            extra = not diff['fullpath'].startswith(self.prefix.rstrip("/"))
+            if keyInfo['from'] == 'None':
+                keyInfo['from'] = None
 
-            if not extra:
-                diff['path'] = diff['fullpath'][len(self.prefix):]
-            else:
-                diff['path'] = diff['fullpath']
+            path = self._relativePath("/" + keyInfo['fullpath'])
 
-            if not diff['path']:
-                diff['path'] = "."
+            if path is None:
+                continue
 
-            self.vols[diff['to']] = {
-                'uuid': diff['to'],
-                'path': diff['path'],
-                'fullpath': diff['fullpath'],
-                'extra': extra,
-                'exclusiveSize': key.size,
-            }
+            diff = Store.Diff(self, keyInfo['to'], keyInfo['from'], key.size)
 
-            self.diffs.append({
-                'from': diff['from'],
-                'to': diff['to'],
-                'size': key.size,
-                'path': diff['path']
-            })
+            logger.debug("Adding %s in %s", diff, path)
 
-        logger.debug("Diffs:\n%s", pprint.pformat(self.diffs))
+            self.diffs[diff.fromVol].append(diff)
+            self.paths[diff.toVol].add(path)
+
+        # logger.debug("Diffs:\n%s", pprint.pformat(self.diffs))
         # logger.debug("Vols:\n%s", pprint.pformat(self.vols))
 
-    def listVolumes(self):
-        """ Return list of volumes that are available. """
-        return [vol for vol in self.vols.values() if not vol['extra']]
+    def getEdges(self, fromVol):
+        """ Return the edges available from fromVol. """
+        return self.diffs[fromVol]
 
-    def getVolume(self, uuid):
-        """ Return info about volume. """
-        return self.vols[uuid]
-
-    def iterEdges(self, fromVol):
-        """ Return the edges available from fromVol.
-
-        Returned edge is a dict: 'to' UUID, estimated 'size' in bytes
-        """
-        for diff in self.diffs:
-            if diff['from'] == fromVol:
-                yield {'to': diff['to'], 'size': diff['size']}
-
-    def hasEdge(self, toUUID, fromUUID):
+    def hasEdge(self, diff):
         """ Test whether edge is in this sink. """
-        for diff in self.diffs:
-            if diff['from'] == fromUUID and diff['to'] == toUUID:
-                return True
-        return False
+        return diff.toVol in [d.toVol for d in self.diffs[diff.fromVol]]
 
-    def receive(self, toUUID, fromUUID, path):
-        """ Return a file-like (stream) object to store a diff. """
-        path = os.path.normpath(os.path.join(self.prefix, path))
-        return _Uploader(self.bucket, self._keyName(toUUID, fromUUID, path))
+    def receive(self, diff, paths, dryrun=False):
+        """ Return Context Manager for a file-like (stream) object to store a diff. """
+        path = self.selectReceivePath(paths)
+        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)
+
+        if Store.skipDryRun(logger, dryrun)("receive %s in %s", keyName, self):
+            return None
+
+        return _Uploader(self.bucket, keyName)
+
+    def receiveVolumeInfo(self, paths, dryrun=False):
+        """ Return Context Manager for a file-like (stream) object to store volume info. """
+        path = self.selectReceivePath(paths)
+        path = path + ".bs"
+
+        if Store.skipDryRun(logger, dryrun)("receive info in '%s'", path):
+            return None
+
+        return _Uploader(self.bucket, path)
 
     theKeyPattern = "^(?P<fullpath>.*)/(?P<to>[-a-zA-Z0-9]*)_(?P<from>[-a-zA-Z0-9]*)$"
 
@@ -157,13 +145,27 @@ class S3Store(Store.Store):
         return "%s/%s_%s" % (path, toUUID, fromUUID)
 
     def _parseKeyName(self, name):
-        match = self.keyPattern.match(name)
-        return match.groupdict() if match else None
+        """ Returns dict with fullpath, to, from. """
+        if name.endswith(".bs"):
+            return {'type': 'info'}
 
-    def send(self, toUUID, fromUUID, streamContext, progress=True):
-        """ Write the diff to the streamContext. """
-        path = self.vols[toUUID]['fullpath']
-        key = self.bucket.get_key(self._keyName(toUUID, fromUUID, path))
+        match = self.keyPattern.match(name)
+        if not match:
+            return None
+
+        match = match.groupdict()
+        match.update(type='diff')
+
+        return match
+
+    def send(self, diff, streamContext, progress=True, dryrun=False):
+        """ Write the diff (toVol from fromVol) to the stream context manager. """
+        path = self.getSendPath(diff.toVol)
+        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)
+        key = self.bucket.get_key(keyName)
+
+        if Store.skipDryRun(logger, dryrun)("send %s in %s", keyName, self):
+            return
 
         with streamContext as stream:
             if progress:
@@ -171,7 +173,7 @@ class S3Store(Store.Store):
             else:
                 key.get_contents_to_file(stream)
 
-        if sys.stdout.isatty():
+        if progress:
             sys.stdout.write("\n")
 
     def _upload(self, stream, keyName):
@@ -191,6 +193,7 @@ class S3Store(Store.Store):
 
 
 class _DisplayProgress:
+
     def __init__(self):
         self.startTime = datetime.datetime.now()
 
@@ -199,7 +202,7 @@ class _DisplayProgress:
             return
 
         elapsed = datetime.datetime.now() - self.startTime
-        mbps = (sent * 8 / (10**6) / elapsed.total_seconds())
+        mbps = (sent * 8 / (10 ** 6) / elapsed.total_seconds())
 
         sys.stdout.write(
             "\r %s: Sent %s of %s (%d%%) (%.3g Mbps) %20s\r" % (
@@ -209,7 +212,7 @@ class _DisplayProgress:
                 int(100 * sent / total),
                 mbps,
                 " ",
-                )
+            )
         )
 
         sys.stdout.flush()
@@ -239,6 +242,10 @@ class _Uploader:
 
     def __exit__(self, exceptionType, exceptionValue, traceback):
         self.close(abort=exceptionType is not None)
+        if exceptionType is not None:
+            logger.error("abort")
+        else:
+            logger.debug("close")
         return False  # Don't supress exception
 
     def write(self, bytes):
